/*
 *  Copyright (C) 2017-2018 Team Kodi
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#pragma once

// Default controller IDs
#define DEFAULT_CONTROLLER_ID "game.controller.default"
#define DEFAULT_KEYBOARD_ID "game.controller.keyboard"
#define DEFAULT_MOUSE_ID "game.controller.mouse"
#define DEFAULT_REMOTE_ID "game.controller.remote"
