/*
 *  Copyright (C) 2016-2018 Team Kodi
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#include "GameResource.h"

#include <utility>

using namespace ADDON;

CGameResource::CGameResource(const AddonInfoPtr& addonInfo) :
  CResource(addonInfo, ADDON_RESOURCE_GAMES)
{
}
