/*
 *  Copyright (C) 2005-2020 Team Kodi
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#pragma once

namespace ADDON
{

/**
 * Struct - RepoInfo
 */
struct RepoInfo
{
  std::string m_repoId;
  std::string m_origin;
};

}; /* namespace ADDON */
