/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class com_plutinosoft_platinum_UPnP */

#ifndef _Included_com_plutinosoft_platinum_UPnP
#define _Included_com_plutinosoft_platinum_UPnP
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     com_plutinosoft_platinum_UPnP
 * Method:    _init
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_com_plutinosoft_platinum_UPnP__1init
  (JNIEnv *, jclass);

/*
 * Class:     com_plutinosoft_platinum_UPnP
 * Method:    _start
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_com_plutinosoft_platinum_UPnP__1start
  (JNIEnv *, jclass, jlong);

/*
 * Class:     com_plutinosoft_platinum_UPnP
 * Method:    _stop
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_com_plutinosoft_platinum_UPnP__1stop
  (JNIEnv *, jclass, jlong);

#ifdef __cplusplus
}
#endif
#endif
