/*
 *  Copyright (C) 2017-2018 Team Kodi
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#pragma once

#define BLUE_BAR                          0
#define LABEL_ROW1                       10
#define LABEL_ROW2                       11
#define LABEL_ROW3                       12

 // Displays current position, visible after seek or when forced
 // Alt, use conditional visibility Player.DisplayAfterSeek
#define LABEL_CURRENT_TIME               22

 // Displays when video is rebuffering
 // Alt, use conditional visibility Player.IsCaching
#define LABEL_BUFFERING                  24

 // Progressbar used for buffering status and after seeking
#define CONTROL_PROGRESS                 23
