/*
 *  Copyright (C) 2007-2018 Team Kodi
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#pragma once

enum EShaderFormat
{
  SHADER_NONE,
  SHADER_YV12,
  SHADER_YV12_9,
  SHADER_YV12_10,
  SHADER_YV12_12,
  SHADER_YV12_14,
  SHADER_YV12_16,
  SHADER_NV12,
  SHADER_YUY2,
  SHADER_UYVY,
  SHADER_NV12_RRG
};
