/*
 *  Copyright (C) 2015-2018 Team Kodi
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#pragma once

// XML definitions
#define LAYOUT_XML_ROOT "layout"
#define LAYOUT_XML_ELM_CATEGORY "category"
#define LAYOUT_XML_ELM_BUTTON "button"
#define LAYOUT_XML_ELM_ANALOG_STICK "analogstick"
#define LAYOUT_XML_ELM_ACCELEROMETER "accelerometer"
#define LAYOUT_XML_ELM_MOTOR "motor"
#define LAYOUT_XML_ELM_RELPOINTER "relpointer"
#define LAYOUT_XML_ELM_ABSPOINTER "abspointer"
#define LAYOUT_XML_ELM_WHEEL "wheel"
#define LAYOUT_XML_ELM_THROTTLE "throttle"
#define LAYOUT_XML_ELM_KEY "key"
#define LAYOUT_XML_ELM_TOPOLOGY "physicaltopology"
#define LAYOUT_XML_ELM_PORT "port"
#define LAYOUT_XML_ELM_ACCEPTS "accepts"
#define LAYOUT_XML_ATTR_LAYOUT_LABEL "label"
#define LAYOUT_XML_ATTR_LAYOUT_ICON "icon"
#define LAYOUT_XML_ATTR_LAYOUT_IMAGE "image"
#define LAYOUT_XML_ATTR_CATEGORY_NAME "name"
#define LAYOUT_XML_ATTR_CATEGORY_LABEL "label"
#define LAYOUT_XML_ATTR_FEATURE_NAME "name"
#define LAYOUT_XML_ATTR_FEATURE_LABEL "label"
#define LAYOUT_XML_ATTR_INPUT_TYPE "type"
#define LAYOUT_XML_ATTR_KEY_SYMBOL "symbol"
#define LAYOUT_XML_ATTR_PROVIDES_INPUT "providesinput"
#define LAYOUT_XML_ATTR_PORT_ID "id"
#define LAYOUT_XML_ATTR_CONTROLLER "controller"

// Controller definitions
#define FEATURE_CATEGORY_FACE "face"
#define FEATURE_CATEGORY_SHOULDER "shoulder"
#define FEATURE_CATEGORY_TRIGGER "triggers"
#define FEATURE_CATEGORY_ANALOG_STICK "analogsticks"
#define FEATURE_CATEGORY_ACCELEROMETER "accelerometer"
#define FEATURE_CATEGORY_HAPTICS "haptics"
#define FEATURE_CATEGORY_MOUSE_BUTTON "mouse"
#define FEATURE_CATEGORY_POINTER "pointer"
#define FEATURE_CATEGORY_LIGHTGUN "lightgun"
#define FEATURE_CATEGORY_OFFSCREEN "offscreen"
#define FEATURE_CATEGORY_KEY "keys"
#define FEATURE_CATEGORY_KEYPAD "keypad"
#define FEATURE_CATEGORY_HARDWARE "hardware"
#define FEATURE_CATEGORY_WHEEL "wheel"
#define FEATURE_CATEGORY_JOYSTICK "joysticks"
#define FEATURE_CATEGORY_PADDLE "paddles"
