/*
 *  Copyright (C) 2005-2018 Team Kodi
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#pragma once

#include <list>

namespace XBMCAddon
{
  template <class T> class List : public std::list<T>
  {
  public:
    static List<T> nullList;
  };
}
