//
//  AppDelegate.h
//  MobileMediaServer
//
//  Created by Sylvain Rebaud on 3/10/15.
//
//

#import <UIKit/UIKit.h>

@interface AppDelegate : UIResponder <UIApplicationDelegate>

@property (strong, nonatomic) UIWindow *window;


@end

