/*****************************************************************
|
|   Neptune - Trust Anchors
|
|   This file is automatically generated by a script, do not edit!
|
| Copyright (c) 2002-2010, Axiomatic Systems, LLC.
| All rights reserved.
|
| Redistribution and use in source and binary forms, with or without
| modification, are permitted provided that the following conditions are met:
|     * Redistributions of source code must retain the above copyright
|       notice, this list of conditions and the following disclaimer.
|     * Redistributions in binary form must reproduce the above copyright
|       notice, this list of conditions and the following disclaimer in the
|       documentation and/or other materials provided with the distribution.
|     * Neither the name of Axiomatic Systems nor the
|       names of its contributors may be used to endorse or promote products
|       derived from this software without specific prior written permission.
|
| THIS SOFTWARE IS PROVIDED BY AXIOMATIC SYSTEMS ''AS IS'' AND ANY
| EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
| WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
| DISCLAIMED. IN NO EVENT SHALL AXIOMATIC SYSTEMS BE LIABLE FOR ANY
| DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
| (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
| LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
| ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
| (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
| SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
 ****************************************************************/

 /* GTE CyberTrust Root CA */
const unsigned char NptTlsTrustAnchor_Base_0001_Data[510] = {
0x30,0x82,0x01,0xfa,0x30,0x82,0x01,0x63
,0x02,0x02,0x01,0xa3,0x30,0x0d,0x06,0x09
,0x2a,0x86,0x48,0x86,0xf7,0x0d,0x01,0x01
,0x04,0x05,0x00,0x30,0x45,0x31,0x0b,0x30
,0x09,0x06,0x03,0x55,0x04,0x06,0x13,0x02
,0x55,0x53,0x31,0x18,0x30,0x16,0x06,0x03
,0x55,0x04,0x0a,0x13,0x0f,0x47,0x54,0x45
,0x20,0x43,0x6f,0x72,0x70,0x6f,0x72,0x61
,0x74,0x69,0x6f,0x6e,0x31,0x1c,0x30,0x1a
,0x06,0x03,0x55,0x04,0x03,0x13,0x13,0x47
,0x54,0x45,0x20,0x43,0x79,0x62,0x65,0x72
,0x54,0x72,0x75,0x73,0x74,0x20,0x52,0x6f
,0x6f,0x74,0x30,0x1e,0x17,0x0d,0x39,0x36
,0x30,0x32,0x32,0x33,0x32,0x33,0x30,0x31
,0x30,0x30,0x5a,0x17,0x0d,0x30,0x36,0x30
,0x32,0x32,0x33,0x32,0x33,0x35,0x39,0x30
,0x30,0x5a,0x30,0x45,0x31,0x0b,0x30,0x09
,0x06,0x03,0x55,0x04,0x06,0x13,0x02,0x55
,0x53,0x31,0x18,0x30,0x16,0x06,0x03,0x55
,0x04,0x0a,0x13,0x0f,0x47,0x54,0x45,0x20
,0x43,0x6f,0x72,0x70,0x6f,0x72,0x61,0x74
,0x69,0x6f,0x6e,0x31,0x1c,0x30,0x1a,0x06
,0x03,0x55,0x04,0x03,0x13,0x13,0x47,0x54
,0x45,0x20,0x43,0x79,0x62,0x65,0x72,0x54
,0x72,0x75,0x73,0x74,0x20,0x52,0x6f,0x6f
,0x74,0x30,0x81,0x9f,0x30,0x0d,0x06,0x09
,0x2a,0x86,0x48,0x86,0xf7,0x0d,0x01,0x01
,0x01,0x05,0x00,0x03,0x81,0x8d,0x00,0x30
,0x81,0x89,0x02,0x81,0x81,0x00,0xb8,0xe6
,0x4f,0xba,0xdb,0x98,0x7c,0x71,0x7c,0xaf
,0x44,0xb7,0xd3,0x0f,0x46,0xd9,0x64,0xe5
,0x93,0xc1,0x42,0x8e,0xc7,0xba,0x49,0x8d
,0x35,0x2d,0x7a,0xe7,0x8b,0xbd,0xe5,0x05
,0x31,0x59,0xc6,0xb1,0x2f,0x0a,0x0c,0xfb
,0x9f,0xa7,0x3f,0xa2,0x09,0x66,0x84,0x56
,0x1e,0x37,0x29,0x1b,0x87,0xe9,0x7e,0x0c
,0xca,0x9a,0x9f,0xa5,0x7f,0xf5,0x15,0x94
,0xa3,0xd5,0xa2,0x46,0x82,0xd8,0x68,0x4c
,0xd1,0x37,0x15,0x06,0x68,0xaf,0xbd,0xf8
,0xb0,0xb3,0xf0,0x29,0xf5,0x95,0x5a,0x09
,0x16,0x61,0x77,0x0a,0x22,0x25,0xd4,0x4f
,0x45,0xaa,0xc7,0xbd,0xe5,0x96,0xdf,0xf9
,0xd4,0xa8,0x8e,0x42,0xcc,0x24,0xc0,0x1e
,0x91,0x27,0x4a,0xb5,0x6d,0x06,0x80,0x63
,0x39,0xc4,0xa2,0x5e,0x38,0x03,0x02,0x03
,0x01,0x00,0x01,0x30,0x0d,0x06,0x09,0x2a
,0x86,0x48,0x86,0xf7,0x0d,0x01,0x01,0x04
,0x05,0x00,0x03,0x81,0x81,0x00,0x12,0xb3
,0x75,0xc6,0x5f,0x1d,0xe1,0x61,0x55,0x80
,0x00,0xd4,0x81,0x4b,0x7b,0x31,0x0f,0x23
,0x63,0xe7,0x3d,0xf3,0x03,0xf9,0xf4,0x36
,0xa8,0xbb,0xd9,0xe3,0xa5,0x97,0x4d,0xea
,0x2b,0x29,0xe0,0xd6,0x6a,0x73,0x81,0xe6
,0xc0,0x89,0xa3,0xd3,0xf1,0xe0,0xa5,0xa5
,0x22,0x37,0x9a,0x63,0xc2,0x48,0x20,0xb4
,0xdb,0x72,0xe3,0xc8,0xf6,0xd9,0x7c,0xbe
,0xb1,0xaf,0x53,0xda,0x14,0xb4,0x21,0xb8
,0xd6,0xd5,0x96,0xe3,0xfe,0x4e,0x0c,0x59
,0x62,0xb6,0x9a,0x4a,0xf9,0x42,0xdd,0x8c
,0x6f,0x81,0xa9,0x71,0xff,0xf4,0x0a,0x72
,0x6d,0x6d,0x44,0x0e,0x9d,0xf3,0x74,0x74
,0xa8,0xd5,0x34,0x49,0xe9,0x5e,0x9e,0xe9
,0xb4,0x7a,0xe1,0xe5,0x5a,0x1f,0x84,0x30
,0x9c,0xd3,0x9f,0xa5,0x25,0xd8};
const unsigned int  NptTlsTrustAnchor_Base_0001_Size = 510;
