/*
 *  Copyright (C) 2000, 2001 Martin Norbäck, Håkan Hjort
 *
 *  This file is part of libdvdnav, a DVD navigation library. It is modified
 *  from a file originally part of the Ogle DVD player.
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#pragma once

void vm_print_mnemonic(vm_cmd_t *command);
void vm_print_cmd(int row, vm_cmd_t *command);

