/*
 *  Copyright (C) 2017-2018 Team Kodi
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  See LICENSES/README.md for more information.
 */

#pragma once

// Analog sticks on the default controller
#define DEFAULT_LEFT_STICK_NAME "leftstick"
#define DEFAULT_RIGHT_STICK_NAME "rightstick"
